"""
The Jacobian Inverse Transposed for axiparallel cube grids is diagonal.
We enforce this through a post-processing transformation (instead of hacking it into UFL)
"""

from dune.perftool.ufl.transformations import ufl_transformation
from ufl.algorithms import MultiFunction
from ufl.classes import (Indexed,
                         JacobianInverse,
                         MultiIndex,
                         Product,
                         Restricted,
                         )


class DiagonalJIT(MultiFunction):
    def __init__(self):
        self.replace = {}
        MultiFunction.__init__(self)

    def expr(self, o):
        return self.reuse_if_untouched(o, *tuple(self(op) for op in o.ufl_operands))

    def multi_index(self, o):
        return MultiIndex(tuple(self.replace.get(i, i) for i in o))

    def index_sum(self, o):
        e, i = o.ufl_operands
        if isinstance(e, Product):
            p1, p2 = e.ufl_operands
            if isinstance(p1, Indexed):
                p, ii = p1.ufl_operands
                if isinstance(p, Restricted):
                    p = p.ufl_operands[0]
                if isinstance(p, JacobianInverse):
                    assert(i[0] == ii[0])
                    self.replace[ii[0]] = ii[1]
                    ret = self(e)
                    self.replace = {}
                    return ret

        return self.reuse_if_untouched(o, self(e), i)


@ufl_transformation(name='axiparallel')
def diagonal_jacobian(e):
    """
    This transformations can generically be described as:
    \sum_k J^{(+/-)}_{k,i}x_k ==> J^{(+/-)}_{i,i}x_i
    """
    return DiagonalJIT()(e)
